"""
This project exposes a Command Line Interface (CLI) that helps with a few of the commonly needed/used tasks in natural language processing.
It helps you do the following:

- Download Amazon Reviews (i.e. from Stanford Snap)
- List Available Amazon Reviews Categories (also from Stanford Snap)
- Mount your Google Drive to your Colab project
- Precompute BERT Embeddings (given sentences are already generated)
- Precompute Sentences from Reviews
- Precompute Labels (i.e. goes over labels and generate the Positive/Negative/Neutral one-hot encodings)

```
$ tum --help

Usage: tum [OPTIONS] COMMAND [ARGS]...

Options:
  --help  Show this message and exit.

Commands:
  download-amazon-reviews
  list-amazon-categories
  mount-google-drive
  precompute-bert
  precompute-labels
  precompute-sentences
```

## Listing Amazon Reviews Categories
```
$ tum list-amazon-categories
```

<img src="https://i.ibb.co/zRhbbvs/Screen-Shot-2019-07-16-at-10-05-13.png" alt="Screen-Shot-2019-07-16-at-10-05-13" border="0">


## Downloading Amazon Reviews
**DISCLAIMER:** We assume that you have proper permissions from Stanford SNAP to download and use the Amazon reviews datasets.

```
$ tum download-amazon-reviews -c Arts
```

<img src="https://i.ibb.co/nzLymRk/Screen-Shot-2019-07-16-at-10-08-05.png" alt="Screen-Shot-2019-07-16-at-10-08-05" border="0">

The rest of the commands are used in a similar manner. You could also use the `--help` on all of the commands exposed with our CLI.
"""

import os

import click
from pyfiglet import figlet_format

from tum.helpers import log
from tum.nlp import preprocessing
from tum.nlp.amazon_reviews_dataset import AMAZON_REVIEWS_CATEGORIES


@click.group()
def cli():
    click.secho(figlet_format("NLP", font="slant"), fg="cyan")


@cli.command()
@click.option("--category", "-c", required=True)
def precompute_labels(category):
    preprocessing.precompute_labels(category)


@cli.command()
@click.option("--category", "-c", required=True)
def precompute_sentences(category):
    preprocessing.precompute_sentences(category)


@cli.command()
@click.option("--category", "-c", required=True)
@click.option("--use-gpu", is_flag=True)
@click.option("--processes", default=1)
def precompute_bert(category, use_gpu, processes):
    preprocessing.precompute_bert(category, use_gpu, processes)


@cli.command()
def mount_google_drive():
    from google.colab import drive

    drive.mount("/content/gdrive")


@cli.command()
def list_amazon_categories():
    log("\n".join(AMAZON_REVIEWS_CATEGORIES))


@cli.command()
@click.option("--category", "-c", required=True)
def download_amazon_reviews(category):
    if category == "finefoods":
        link = "https://snap.stanford.edu/data/finefoods.txt.gz"
    else:
        link = f"https://snap.stanford.edu/data/amazon/{category}.txt.gz"
    os.system(f"wget {link}")
    os.system(f"gzip -d {category}.txt.gz")


if __name__ == "__main__":
    cli()
