import torch
import torch.nn as nn
import torch.nn.functional as F

import numpy as np
from sklearn import metrics

import torch.utils.data as data_utils
import torch.optim as optim
from torch.autograd import Variable
from torch.utils.tensorboard import SummaryWriter


class Attention(nn.Module):
    def __init__(self, learning_rate=0.0005, weight_decay=10e-5,
                 loss=torch.nn.CrossEntropyLoss(),
                 optimizer='adam'):
        super(Attention, self).__init__()
        self.L = 500
        self.D = 128
        self.K = 1
        self.writer = SummaryWriter()

        if torch.cuda.is_available():
            print('GPU is ON!')
            self.cuda_available = True
        else:
            self.cuda_available = False

        self.feature_extractor_part1 = nn.Sequential(
            nn.Conv1d(768, 20, kernel_size=1),
            nn.ReLU(),
            nn.MaxPool2d(2, stride=2),
            nn.Conv1d(10, 50, kernel_size=1),
            nn.ReLU(),
            nn.MaxPool2d(2, stride=2)
        )

        self.feature_extractor_part2 = nn.Sequential(
            nn.Linear(25 * 1, self.L),
            nn.ReLU(),
        )

        self.attention = nn.Sequential(
            nn.Linear(self.L, self.D),
            nn.Tanh(),
            nn.Linear(self.D, self.K)
        )

        self.classifier = nn.Sequential(
            nn.Linear(self.L * self.K, 3),
            nn.Softmax()
        )

        self.loss = loss
        if optimizer == 'adam':
            optimizer = optim.Adam(self.parameters(), lr=learning_rate, betas=(0.9, 0.999), weight_decay=weight_decay)
        elif optimizer == 'sgd':
            optimizer = optim.SGD(self.parameters(), lr=learning_rate, weight_decay=weight_decay, momentum=0.9)

        self.optimizer = optimizer

    def forward(self, x):
        H = self.feature_extractor_part1(x)
        H = H.view(-1, 25 * 1)
        H = self.feature_extractor_part2(H)  # NxL

        A = self.attention(H)  # NxK
        A = torch.transpose(A, 1, 0)  # KxN
        A = F.softmax(A, dim=1)  # softmax over N

        M = torch.mm(A, H)  # KxL

        Y_prob = self.classifier(M)
        Y_hat = torch.argmax(Y_prob, dim=1).float()

        return Y_prob, Y_hat, A

    def train_network(self, epochs, embeddings, labels,  validation_embeddings=None, validation_labels=None, save_checkpoint=False, checkpoint_filename='checkpoint', do_validation=False):

        self.train()
        for epoch in range(epochs + 1):
            for i, (data, label) in enumerate(zip(embeddings, labels)):
                data = torch.from_numpy(data)
                bag_label = label

                if self.cuda_available:
                    data, bag_label = data.cuda(), torch.tensor(bag_label).cuda()

                d = data.reshape(*list(data.shape), 1)
                d = torch.tensor(d)
                one_hot = F.one_hot(torch.tensor(bag_label), 3)
                one_hot = one_hot.unsqueeze(0)

                data, bag_label = Variable(d), Variable(one_hot)

                self.optimizer.zero_grad()

                # Forward pass, backward pass, optimize
                prob, hat, A = self.forward(data)

                loss_size = self.loss(prob, torch.argmax(bag_label, 1))
                loss_size.backward()
                self.optimizer.step()

                accuracy = (torch.argmax(bag_label, 1).to(torch.float) == hat.squeeze()).float().sum()

                if (i + 1) % 10000 == 0:
                    print('Epoch [{}/{}], Step [{}/{}], Loss: {:.4f}, Acc: {:.2f}'
                          .format(epoch, epochs, i + 1, len(embeddings), loss_size.item(), accuracy.item()))

                    # 1. Log scalar values (scalar summary)
                    info = {'loss': loss_size.item(), 'accuracy': accuracy.item()}
                    for tag, value in info.items():
                        self.writer.add_scalar(tag, value, i + 1)
                    # 2. Log values and gradients of the parameters (histogram summary)
                    for tag, value in self.named_parameters():
                        tag = tag.replace('.', '/')
                        self.writer.add_histogram(tag, value.data.cpu().numpy(), i + 1)
                        self.writer.add_histogram(tag + '/grad', value.grad.data.cpu().numpy(), i + 1)

            if save_checkpoint:
                self.save_checkpoint(checkpoint_filename, epoch)

            if do_validation:
                self.validate(validation_embeddings, validation_labels)

    def test_network(self, embeddings, labels):
        self.eval()
        correct = 0
        y_target = list()
        y_pred = list()
        for i, (data, label) in enumerate(zip(embeddings, labels)):
            data = torch.from_numpy(data)
            bag_label = label
            if self.cuda_available:
                data, bag_label = data.cuda(), torch.tensor(bag_label).cuda()  # update
            d = data.reshape(*list(data.shape), 1)
            d = torch.tensor(d)
            one_hot = F.one_hot(torch.tensor(bag_label), 3)
            one_hot = one_hot.unsqueeze(0)

            data, bag_label = Variable(d), Variable(one_hot)

            prob, hat, A = self.forward(data)
            y_target.append(int(label))
            y_pred.append(hat.int().item())

            if int(hat.item()) == label:
                correct += 1

        return y_target, y_pred

    def evaluate(self, y_target, y_pred):
        acc = metrics.accuracy_score(y_target, y_pred)
        f1_micro = metrics.f1_score(y_target, y_pred, average="micro")
        f1_macro = metrics.f1_score(y_target, y_pred, average="macro")
        f1_weighted = metrics.f1_score(y_target, y_pred, average="weighted")

        print((
            f"Accuracy: {acc}; \n"
            f"F1-Score (Micro): {f1_micro}; \n"
            f"F1-Score (Macro): {f1_macro}; \n"
            f"F1-Score (Weighted): {f1_weighted}"))

    def validate(self, embeddings, labels):
        correct = 0
        y_target = list()
        y_pred = list()
        for i, (data, label) in enumerate(zip(embeddings, labels)):
            data = torch.from_numpy(data)
            bag_label = label
            if self.cuda_available:
                data, bag_label = data.cuda(), torch.tensor(bag_label).cuda()
            d = data.reshape(*list(data.shape), 1)
            one_hot = F.one_hot(torch.tensor(bag_label), 3)
            data, bag_label = Variable(torch.tensor(d)), Variable(one_hot.unsqueeze(0))

            prob, hat, A = self.forward(data)
            y_target.append(int(label))
            y_pred.append(hat.int().item())
            if int(hat.item()) == label:
                correct += 1

        print('Validation Accuracy : ', (correct / len(labels)) * 100)

    def analyze(self, embeddings, labels, num_samples=50):
        for i, (data, label) in enumerate(zip(embeddings, labels)):
            data = torch.from_numpy(data)  # add
            bag_label = label
            if self.cuda_available:
                data, bag_label = data.cuda(), torch.tensor(bag_label).cuda()  # update
            d = data.reshape(*list(data.shape), 1)
            d = torch.tensor(d)
            one_hot = F.one_hot(torch.tensor(bag_label), 3)
            one_hot = one_hot.unsqueeze(0)

            data, bag_label = Variable(d), Variable(one_hot)
            prob, hat, A = self.forward(data)
            print(i)
            print('Prob :', prob, '\nPredicted : ', hat, '\nAttention : ', A, '\nTarget : ', label, '\n*************************\n')
            i += 1
            if i == num_samples:
                break

    def tensorboard_logging(self, accuracy, i, loss_size):
        # 1. Log scalar values (scalar summary)
        info = {'loss': loss_size.item(), 'accuracy': accuracy.item()}
        for tag, value in info.items():
            self.writer.add_scalar(tag, value, i + 1)
        # 2. Log values and gradients of the parameters (histogram summary)
        for tag, value in self.named_parameters():
            tag = tag.replace('.', '/')
            self.writer.add_histogram(tag, value.data.cpu().numpy(), i + 1)
            self.writer.add_histogram(tag + '/grad', value.grad.data.cpu().numpy(), i + 1)

    def save_checkpoint(self, filename, epoch):
        torch.save({
            'epoch': epoch,
            'model_state_dict': self.state_dict(),
            'optimizer_state_dict': self.optimizer.state_dict(),
            'loss': self.loss,
        }, filename + '_epoch_' + str(epoch) + '.pt')

    def load_checkpoint(self, checkpoint):
        checkpoint = torch.load(checkpoint)
        self.load_state_dict(checkpoint['model_state_dict'])
        self.optimizer.load_state_dict(checkpoint['optimizer_state_dict'])

        return checkpoint['epoch'], checkpoint['loss']

    def resume_training_network(self, checkpoint, epochs, embeddings, labels,  validation_embeddings=None, validation_labels=None, save_checkpoint=False, checkpoint_filename='checkpoint', do_validation=False):
        start_epoch, last_loss = self.load_checkpoint(checkpoint)
        self.train()
        for epoch in range(start_epoch + 1, epochs + 1):
            for i, (data, label) in enumerate(zip(embeddings, labels)):
                data = torch.from_numpy(data)
                bag_label = label

                if self.cuda_available:
                    data, bag_label = data.cuda(), torch.tensor(bag_label).cuda()

                d = data.reshape(*list(data.shape), 1)
                d = torch.tensor(d)
                one_hot = F.one_hot(torch.tensor(bag_label), 3)
                one_hot = one_hot.unsqueeze(0)

                data, bag_label = Variable(d), Variable(one_hot)

                self.optimizer.zero_grad()

                # Forward pass, backward pass, optimize
                prob, hat, A = self.forward(data)

                loss_size = self.loss(prob, torch.argmax(bag_label, 1))
                loss_size.backward()
                self.optimizer.step()

                accuracy = (torch.argmax(bag_label, 1).to(torch.float) == hat.squeeze()).float().sum()

                if (i + 1) % 10000 == 0:
                    print('Epoch [{}/{}], Step [{}/{}], Loss: {:.4f}, Acc: {:.2f}'
                          .format(epoch, epochs, i + 1, len(embeddings), loss_size.item(), accuracy.item()))

                    # 1. Log scalar values (scalar summary)
                    info = {'loss': loss_size.item(), 'accuracy': accuracy.item()}
                    for tag, value in info.items():
                        self.writer.add_scalar(tag, value, i + 1)
                    # 2. Log values and gradients of the parameters (histogram summary)
                    for tag, value in self.named_parameters():
                        tag = tag.replace('.', '/')
                        self.writer.add_histogram(tag, value.data.cpu().numpy(), i + 1)
                        self.writer.add_histogram(tag + '/grad', value.grad.data.cpu().numpy(), i + 1)

            if save_checkpoint:
                self.save_checkpoint(checkpoint_filename, epoch)

            if do_validation:
                self.validate(validation_embeddings, validation_labels)
