from glob import glob
from typing import NamedTuple, Iterable, Optional
import os


SCORE_PREFIX = "review/score: "
SUMMARY_PREFIX = "review/summary: "
TEXT_PREFIX = "review/text: "
AMAZON_REVIEWS_CATEGORIES = [
    "Amazon_Instant_Video",
    "Arts",
    "Automotive",
    "Baby",
    "Beauty",
    "Books",
    "Cell_Phones_&_Accessories",
    "Clothing_&_Accessories",
    "Electronics",
    "Furniture",
    "Gift_Cards",
    "Gourmet_Foods",
    "Health",
    "Home_&_Kitchen",
    "Industrial_&_Scientific",
    "Jewelry",
    "Kindle_Store",
    "Movies_&_TV",
    "Music",
    "Musical_Instruments",
    "Office_Products",
    "Patio",
    "Pet_Supplies",
    "Shoes",
    "Software",
    "Sports_&_Outdoors",
    "Tools_&_Home_Improvement",
    "Toys_&_Games",
    "Video_Games",
    "Watches",
    "finefoods",
]


class Review(NamedTuple):
    """Wrapper for reviews.

    Args:
        score: float
        summary: string
        text: string
    """

    score: float
    summary: str
    text: str


def read_reviews_file(path: str) -> Iterable[Review]:
    """Reads reviews file (from Stanford SNAP `*.txt` format)

    Args:
        path: string. The reviews text file path; Should end with `.txt`.

    Returns:
        Generator that yields `Review` objects.

    Examples:
        ```
        review/score: 5.0
        review/summary: Great! But please read on before purchasing ext batteries!
        review/text: This is simpily the best rechargable ...
        ```
    """
    score, summary, text = None, None, None

    encoding = os.environ.get("REVIEWS_ENCODING", "latin-1")

    with open(path, "rt", encoding=encoding) as f:
        for line in f:
            score = extract_with_prefix(SCORE_PREFIX, line) or score
            summary = extract_with_prefix(SUMMARY_PREFIX, line) or summary
            text = extract_with_prefix(TEXT_PREFIX, line) or text

            if score and summary and text:
                yield Review(score=float(score), summary=summary, text=text)
                score, summary, text = None, None, None


def extract_with_prefix(prefix: str, line: str) -> Optional[str]:
    if prefix in line:
        return line.replace(prefix, "").rstrip()


def batch(iterable, n=1):
    length = len(iterable)
    for ndx in range(0, length, n):
        yield iterable[ndx : min(ndx + n, length)]


__all__ = ["read_reviews_file", "Review"]
