import joblib
import numpy as np
import xgboost as xgb
from keras import callbacks, layers, models
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB
from sklearn.svm import SVC

from tum.nlp.layers import Attention
from tum.nlp.preprocessing import BERT_SIZE
from tum.nlp.training import (
    TestResult,
    average_all_sentences_per_review,
    count_equal_classes,
    data_generator,
    data_generator_with_fixed_batch_size,
    get_test_result,
    test_model,
)


class ITrainableModel:
    """ITrainableModel

    All trainable models should define: train, test, save, and load functions!
    """

    def __init__(self, *args, **kwargs):
        raise NotImplementedError()

    def train(
        self,
        X_train: np.ndarray,
        y_train: np.ndarray,
        X_test: np.ndarray,
        y_test: np.ndarray,
    ):
        """Trains the model defined by interface children.

        This is an interface function that required all children
        to provide their training immplementation.

        Args:
            X_train: numpy array `List[List[np.ndarray]]`

                ```
                X_train = np.array([
                    [np.array(...), np.array(...), ...],
                    ...,
                    [np.array(...), np.array(...), ...]
                ])
                ```

                - `X_train` corresponds to the number of reviews.
                - `X_train[0]` corresponds to the embeddings for sentences in the first review.
                - `X_train[0][0]` corresponds to the embedding vector for first sentence of first review.
            y_train: np.ndarray
            X_test: np.ndarray
            y_test: np.ndarray
        """
        raise NotImplementedError()

    def test(self, X: np.ndarray, y: np.ndarray):
        raise NotImplementedError()

    def save(self, name: str):
        raise NotImplementedError()

    def load(self, name: str):
        raise NotImplementedError()


class ILSTMModel(ITrainableModel):
    def __init__(self, *args, **kwargs):
        self.model = None

    def test(self, X, y) -> TestResult:
        return test_model(self.model, X, y, equal_classes=True)

    def on_epoch_end(self, X, y):
        def run_test(epoch, logs):
            print(f"Testing epoch {epoch}.")
            self.test(X, y)

        return run_test

    def train(self, X_train, y_train, X_test, y_test):
        self.model.fit_generator(
            generator=data_generator(X_train, y_train, equal_classes=True),
            steps_per_epoch=count_equal_classes(X_train, y_train),
            epochs=25,
            callbacks=[
                callbacks.LambdaCallback(
                    on_epoch_end=lambda epoch, logs: self.on_epoch_end(X_test, y_test)
                )
            ],
        )

    def save(self, name: str):
        assert name.endswith(".h5"), "Weights shall be saved as h5 files!"
        self.model.save_weights(name)

    def load(self, name: str):
        assert name.endswith(".h5"), "Weights shall be saved as h5 files!"
        self.model.load_weights(name)


class LSTMModel(ILSTMModel):
    def __init__(self, *args, **kwargs):
        self.feed = layers.Input(shape=(None, BERT_SIZE))
        out = layers.LSTM(
            25, dropout=0.1, recurrent_dropout=0.1, return_sequences=False
        )(self.feed)
        out = layers.Dense(3, activation="softmax")(out)
        self.model = models.Model(inputs=self.feed, outputs=out)
        self.model.compile(
            loss="categorical_crossentropy", optimizer="adam", metrics=["accuracy"]
        )
        print(self.model.summary())


class TwoLayerLSTMModel(ILSTMModel):
    def __init__(self, *args, **kwargs):
        self.feed = layers.Input(shape=(None, BERT_SIZE))
        l1 = layers.LSTM(25, dropout=0.1, recurrent_dropout=0.1, return_sequences=True)(
            self.feed
        )
        out = layers.LSTM(
            25, dropout=0.1, recurrent_dropout=0.1, return_sequences=False
        )(l1)
        out = layers.Dense(3, activation="softmax")(out)
        self.model = models.Model(inputs=self.feed, outputs=out)
        self.model.compile(
            loss="categorical_crossentropy", optimizer="adam", metrics=["accuracy"]
        )
        print(self.model.summary())


class DeepNeuralModel:
    def __init__(self, *args, **kwargs):
        self.feed = layers.Input(shape=(BERT_SIZE,))
        fc1 = layers.Dense(128, activation="relu")(self.feed)
        d1 = layers.Dropout(0.1)(fc1)
        fc2 = layers.Dense(64, activation="relu")(d1)
        d2 = layers.Dropout(0.1)(fc2)
        out = layers.Dense(3, activation="softmax")(d2)
        self.model = models.Model(inputs=self.feed, outputs=out)
        self.model.compile(
            loss="categorical_crossentropy", optimizer="adam", metrics=["accuracy"]
        )
        print(self.model.summary())

    def test(self, X, y) -> TestResult:
        X_avg = average_all_sentences_per_review(X)

        assert X_avg.shape[-1] == BERT_SIZE
        assert X_avg.shape[0] == y.shape[0]

        return test_model(self.model, X_avg, y, equal_classes=True)

    def on_epoch_end(self, X, y):
        def run_test(epoch, logs):
            print(f"Testing epoch {epoch}.")
            self.test(X, y)

        return run_test

    def train(self, X_train, y_train, X_test, y_test):
        X_train_avg = average_all_sentences_per_review(X_train)

        assert X_train_avg.shape[-1] == BERT_SIZE
        assert X_train_avg.shape[0] == y_train.shape[0]

        self.model.fit_generator(
            generator=data_generator_with_fixed_batch_size(
                X_train_avg, y_train, equal_classes=True
            ),
            steps_per_epoch=count_equal_classes(X_train_avg, y_train),
            epochs=25,
            callbacks=[
                callbacks.LambdaCallback(on_epoch_end=self.on_epoch_end(X_test, y_test))
            ],
        )

    def save(self, name: str):
        assert name.endswith(".h6"), "Weights shall be saved as h5 files!"
        self.model.save_weights(name)

    def load(self, name: str):
        assert name.endswith(".h5"), "Weights shall be saved as h5 files!"
        self.model.load_weights(name)


class BidirectionalLSTMWithAttention(ITrainableModel):
    MAX_LENGTH = 32

    def __init__(self, *args, **kwargs):
        inp = layers.Input(shape=(self.MAX_LENGTH, BERT_SIZE))
        x = layers.Bidirectional(
            layers.LSTM(25, return_sequences=True, dropout=0.25, recurrent_dropout=0.25)
        )(inp)
        x = Attention(self.MAX_LENGTH)(x)
        x = layers.Dense(25, activation="relu")(x)
        x = layers.Dropout(0.25)(x)
        x = layers.Dense(3, activation="softmax")(x)
        self.model = layers.Model(inputs=inp, outputs=x)

    def test(self, X, y) -> TestResult:
        X_avg = average_all_sentences_per_review(X)

        assert X_avg.shape[-1] == BERT_SIZE
        assert X_avg.shape[0] == y.shape[0]

        return test_model(self.model, X_avg, y, equal_classes=True)

    def on_epoch_end(self, X, y):
        def run_test(epoch, logs):
            print(f"Testing epoch {epoch}.")
            self.test(X, y)

        return run_test

    def train(self, X_train, y_train, X_test, y_test):
        X_train_avg = average_all_sentences_per_review(X_train)

        assert X_train_avg.shape[-1] == BERT_SIZE
        assert X_train_avg.shape[0] == y_train.shape[0]

        self.model.fit_generator(
            generator=data_generator_with_fixed_batch_size(
                X_train_avg, y_train, equal_classes=True
            ),
            steps_per_epoch=count_equal_classes(X_train_avg, y_train),
            epochs=25,
            callbacks=[
                callbacks.LambdaCallback(on_epoch_end=self.on_epoch_end(X_test, y_test))
            ],
        )

    def save(self, name: str):
        assert name.endswith(".h6"), "Weights shall be saved as h5 files!"
        self.model.save_weights(name)

    def load(self, name: str):
        assert name.endswith(".h5"), "Weights shall be saved as h5 files!"
        self.model.load_weights(name)


class ISklearnTrainableModel(ITrainableModel):
    def __init__(self):
        self.clf = None

    def train(self, X, y):
        X_averaged = average_all_sentences_per_review(X)
        self.clf.fit(X_averaged, y.argmax(axis=-1))

    def test(self, X, y) -> TestResult:
        X_averaged = average_all_sentences_per_review(X)
        y_pred = self.clf.predict_proba(X_averaged)
        return get_test_result(y, y_pred)

    def save(self, name: str):
        assert name.endswith(
            ".joblib"
        ), "logistic classifier names should end with .joblib"
        joblib.dump(self.clf, name)

    def load(self, name: str):
        assert name.endswith(
            ".joblib"
        ), "logistic classifier names should end with .joblib"
        self.clf = joblib.load(name)


class LogisticRegressionModel(ISklearnTrainableModel):
    def __init__(self):
        self.clf = LogisticRegression(
            random_state=0, solver="lbfgs", multi_class="multinomial"
        )


class XGBoostModel(ISklearnTrainableModel):
    def __init__(self):
        self.clf = xgb.XGBClassifier(
            objective="binary:logistic", random_state=42, n_jobs=16, silent=False
        )


class GaussianClassifierModel(ISklearnTrainableModel):
    def __init__(self):
        self.clf = GaussianNB()


class SvmModel(ISklearnTrainableModel):
    def __init__(self):
        self.clf = SVC(kernel="rbf", probability=True, verbose=True)
