import os
import re
from glob import glob
from tum.nlp.amazon_reviews_dataset import read_reviews_file

import click
import mxnet as mx
import nltk
import numpy as np
from bert_embedding import BertEmbedding
from tqdm import tqdm

BERT_SIZE = 768


def get_data_and_labels(path):
    X = []
    y = []

    for review in read_reviews_file(path):
        X.append(review.text)
        y.append(review.score)

    return X, y


def get_bert_client(use_gpu=False, *args, **kwargs) -> BertEmbedding:
    if not hasattr(get_bert_client, "client"):
        if use_gpu:
            ctx = mx.gpu(0)
            get_bert_client.client = BertEmbedding(ctx=ctx, *args, **kwargs)
        else:
            get_bert_client.client = BertEmbedding(*args, **kwargs)

    return get_bert_client.client


def precompute_labels(category: str):
    click.secho(f"Pre-computing labels on {category} ...", fg="green")

    positive = [1, 0, 0]
    neutral = [0, 1, 0]
    negative = [0, 0, 1]

    label_map = {1: negative, 2: negative, 3: neutral, 4: positive, 5: positive}

    read_path = f"{category}.txt"
    write_path = f"{category}.labels.npy"

    if os.path.exists(write_path):
        print("File exists; Exiting ...")
        return

    X, y = get_data_and_labels(read_path)

    labels = []
    for score in tqdm(y):
        labels.append(label_map[score])

    labels = np.array(labels)
    np.save(write_path, labels)


def precompute_sentences(category):
    click.secho(f"Pre-computing sentences on {category} ...", fg="green")

    nltk.download("punkt")

    read_path = f"{category}.txt"
    write_path = f"{category}.sentences.npy"
    if os.path.exists(write_path):
        print("File exists; Exiting ...")
        return

    X, y = get_data_and_labels(read_path)

    sentences = []
    for review in tqdm(X):
        sentences.append(nltk.sent_tokenize(review))

    sentences = np.array(sentences)
    np.save(write_path, sentences)


def precompute_bert(category, use_gpu, processes):
    click.secho(f"Pre-computing bert embeddings on {category} ...", fg="green")

    read_path = f"{category}.sentences.npy"
    write_path = f"{category}.sentences.bert.npy"

    if os.path.exists(write_path):
        print("Bert embeddings already exist; Exiting ...")
        return

    if not os.path.exists(read_path):
        print("Please, get the sentences for the category first!")
        return

    reviews_sentences = np.load(read_path)

    embeddings = compute_embeddings_sequential(category, reviews_sentences, use_gpu)

    embeddings = np.array(embeddings)
    np.save(write_path, embeddings)


def compute_review_for_sentences(args):
    use_gpu, (index, sentences) = args
    bert = get_bert_client(use_gpu)
    sub_embeddings = bert(sentences)
    return index, [sum(e[1]) / len(e[1]) for e in sub_embeddings]


def extract_chunk_id_and_size(filename):
    # group 0 -> the actual matched string
    # group 1 -> the category
    # group 2 -> the chunk id
    sentence_format = r"(.*?)\.sentences\.bert\.(\d+)\.(\d+).npy"
    result = re.search(sentence_format, filename)
    chunk_id = result.group(2)
    chunk_size = result.group(3)
    return chunk_id, chunk_size


def count_from_chunk(chunk_data):
    chunk_id, chunk_size = chunk_data
    return int(chunk_id) * int(chunk_size)


def highest_checkpoint(category):
    previous_embeddings = glob(f"{category}.sentences.bert.*.*.npy")
    all_chunks = map(extract_chunk_id_and_size, previous_embeddings)
    all_chunks = list(map(count_from_chunk, all_chunks))
    all_chunks.append(0)
    return max(all_chunks)


def compute_embeddings_sequential(category, reviews_sentences, use_gpu):
    total = len(reviews_sentences)
    chunk_size = 1000
    partition_id = 0  # Need to set the default (if all reviews less than one chunk)
    all_embeddings = []
    previous_total = highest_checkpoint(category)
    for index, sentences in tqdm(enumerate(reviews_sentences), total=total):
        if index < previous_total:
            continue

        _, embeddings = compute_review_for_sentences((use_gpu, (index, sentences)))
        all_embeddings.append(embeddings)

        if len(all_embeddings) == chunk_size:
            partition_id = int((index + 1) / chunk_size)
            new_file = f"{category}.sentences.bert.{partition_id}.{chunk_size}.npy"
            np.save(new_file, all_embeddings)
            all_embeddings = []

    if all_embeddings:
        new_file = f"{category}.sentences.bert.{partition_id + 1}.{chunk_size}.npy"
        np.save(new_file, all_embeddings)

    return all_embeddings
