from collections import defaultdict
from typing import Callable, NamedTuple
from typing import List
from tum.nlp.preprocessing import BERT_SIZE

import numpy as np
from keras import models
from sklearn import metrics

CHUNK_SIZE = 1000
CHUNKS_COUNT = 400


class TestResult(NamedTuple):
    accuracy: float
    f1_micro: float
    f1_macro: float
    f1_weighted: float

    def __str__(self):
        return (
            f"Accuracy: {self.accuracy}; \n"
            f"F1-Score (Micro): {self.f1_micro}; \n"
            f"F1-Score (Macro): {self.f1_macro}; \n"
            f"F1-Score (Weighted): {self.f1_weighted}"
        )


class Dataset(NamedTuple):
    X: np.ndarray
    y: np.ndarray


class Data(NamedTuple):
    train: Dataset
    test: Dataset


def group_data_by_label(X: np.ndarray, y: np.ndarray):
    data_per_label = defaultdict(list)
    for index in range(X.shape[0]):
        label = y[index].argmax()
        data_per_label[label].append((X[index], y[index]))
    return data_per_label


def get_equal_classes(X: np.ndarray, y: np.ndarray):
    data_per_label = group_data_by_label(X, y)
    min_label = min([len(v) for v in data_per_label.values()])
    all_X, all_y = [], []

    for data in data_per_label.values():
        np.random.shuffle(data)

    for index in range(min_label):
        for data in data_per_label.values():
            _X, _y = data[index]
            all_X.append(_X)
            all_y.append(_y)

    return np.array(all_X), np.array(all_y)


def count_equal_classes(X: np.ndarray, y: np.ndarray):
    data_per_label = group_data_by_label(X, y)
    return len(data_per_label.values()) * min([len(v) for v in data_per_label.values()])


def data_generator(X, y, equal_classes=False):
    while True:
        if equal_classes:
            _X, _y = get_equal_classes(X, y)
        else:
            _X, _y = X, y

        for index in range(_X.shape[0]):
            yield np.array(_X[index])[np.newaxis, ...], _y[index][np.newaxis, ...]


def data_generator_with_fixed_batch_size(X, y, batch_size=32, equal_classes=False):
    while True:
        if equal_classes:
            _X, _y = get_equal_classes(X, y)
        else:
            _X, _y = X, y

        batches = int(_X.shape[0] / batch_size)

        for index in range(batches):
            yield (
                np.array(_X[index * batch_size : (index + 1) * batch_size]),
                _y[index * batch_size : (index + 1) * batch_size],
            )


def padded_data_generator_with_fixed_batch_size(
    X, y, batch_size=32, max_length=32, equal_classes=False
):
    while True:
        if equal_classes:
            _X, _y = get_equal_classes(X, y)
        else:
            _X, _y = X, y

        batches = int(_X.shape[0] / batch_size)

        for index in range(batches):
            X_batch = pad_reviews(_X[index * batch_size : (index + 1) * batch_size])
            assert X_batch.shape == (batch_size, max_length, BERT_SIZE)
            yield (X_batch, _y[index * batch_size : (index + 1) * batch_size])


def pad_reviews(X: np.ndarray, max_length: int) -> np.ndarray:
    reviews = [pad_review(review, max_length) for review in X]
    return np.array(reviews)


def pad_review(review: List[np.ndarray], max_length: int) -> np.ndarray:
    new_review = np.zeros((max_length, BERT_SIZE))
    new_review[: len(review), ...] = np.array(review)
    return new_review


def test_model(
    model: models.Model, X: np.ndarray, y: np.ndarray, equal_classes: bool = True
) -> TestResult:
    print(f"Testing the model (EqualClasse: {equal_classes}) ...")
    __X, __y = get_equal_classes(X, y)
    y_pred = model.predict_generator(
        data_generator(__X, __y, equal_classes=False), steps=__X.shape[0], verbose=True
    )

    result = get_test_result(y_true=__y, y_pred=y_pred)
    print(result)

    return result


def get_test_result(y_true, y_pred) -> TestResult:
    args = y_true.argmax(axis=1), y_pred.argmax(axis=1)

    return TestResult(
        accuracy=metrics.accuracy_score(*args),
        f1_micro=metrics.f1_score(*args, average="micro"),
        f1_macro=metrics.f1_score(*args, average="macro"),
        f1_weighted=metrics.f1_score(*args, average="weighted"),
    )


def patch_numpy_to_allow_picke_reads():
    """
    This patch solves a problem arising from a recent numpy upgrade. They disabled
    using pickle for storing and loading data; Therefore, we need to set this manually.

    Background:
        https://stackoverflow.com/a/56243777

    Exception:
        --> def open_memmap(filename, mode='r+', dtype=None, shape=None, fortran_order=False, version=None):
        ValueError: Object arrays cannot be loaded when allow_pickle=False
    """
    np_load_old = np.load
    np.load = lambda *args, **kwargs: np_load_old(*args, **kwargs, allow_pickle=True)


def load_precomputed_data(
    total_chunks: int, X_path: Callable[[int], str], y_path: Callable[[], str]
) -> Data:
    X = []
    y = np.load(y_path())

    # Trim labels by total chunks we are loading
    y = y[: total_chunks * CHUNK_SIZE, ...]

    for chunk_id in range(1, total_chunks + 1):
        print(f"Loading chunk # {chunk_id}")
        X += np.load(X_path(chunk_id)).tolist()

    CUTOFF = int(total_chunks * CHUNK_SIZE * 0.75)

    X_train = np.array(X[:CUTOFF])
    y_train = y[:CUTOFF]

    X_test = np.array(X[CUTOFF:])
    y_test = y[CUTOFF:]

    assert (
        X_train.shape[0] == y_train.shape[0]
    ), "Both training X and y should be of same length"
    assert (
        X_test.shape[0] == y_test.shape[0]
    ), "Both testing X and y should be of same length"

    return Data(train=Dataset(X=X_train, y=y_train), test=Dataset(X=X_test, y=y_test))


def average_all_sentences_per_review(X):
    review_vectors = []
    for sentences in X:
        review = np.array(sentences).mean(axis=0)
        review_vectors.append(review)
    return np.array(review_vectors)
